#!/usr/bin/ruby

#############################################################################
# iDaVer.rb (for Integrated DAta model VERifier)
# 
# Author: Jaideep Nijjar, Ivan Boci
# Date: February 2014
# Ruby Version: 1.9.3 
#
# Use this tool to perform verification of data models of Ruby on Rails 2.0
#  web applications.  
# There are two approaches: bounded verification (using Alloy Analyzer) 
#  and unbounded (using an SMT Solver, Z3). 
# Use property templates to specify the properties you wish to check about
#  the data model.
# This tool can also be used to automatically infer properties about your
#  data model.  These properties are then checked using unbounded verification.
#############################################################################


require 'optparse'	# for OptionParser
require 'rbconfig'	# for RUBY_PLATFORM
require 'fileutils'	# for Fileutils.cp method
gem 'activesupport'
require 'active_support/inflector'  # for pluralize method
require_relative 'os' # for executing SPASS correctly depending on OS
require_relative 'csv_hash_formatter' # for outputting in CSV format
require_relative 'smt_ar_processor'   # for creating ERDs
require 'tempfile'
require 'set'


#--------------------------------------------------#
#------------------- Constants --------------------#
#--------------------------------------------------#

ALLOYSPEC = "alloyspec.temp"
SMTSPEC = "smtspec.temp"
INSTANCE = "z3instance.temp"
DEFAULT_ERD_FILENAME = "erd.dot"
DEFAULT_BOUND = 20
RelationInfo = Struct.new(:num_args, :arg_types, :return_type)



#--------------------------------------------------#
#-------------------- Methods ---------------------#
#--------------------------------------------------#

# Extracts properties specified in given 'filename'
def get_properties( filename )
    if not File.file? filename
        puts "Property file not found (#{filename})."
        exit
    end
    
    
    properties = []  # property strings extracted from input file
    l = 0		         # line number in file
    
    IO.foreach(filename) do |line|
        l += 1
        next if line.strip.empty? or line =~ /^\s*\/\//  # skip empty lines and comments
        
        property = nil	  # holds important part of input, as a string: "propName classA classB [classC]"
        static = false	  # true if this is a static property; false otherwise
        predicate = false # true if this is a predicate property; false if an assertion property
        delclass = nil	  # stores the class-to-delete for non-static properties
        
        if line =~ /alwaysRelated\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "alwaysRelated" + " " + $1 + " " + $2
            static = true
        elsif line =~ /someUnrelated\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "someUnrelated" + " " + $1 + " " + $2
            static = true
            predicate = true
        elsif line =~ /multipleRelated\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "multipleRelated" + " " + $1 + " " + $2
            static = true
            predicate = true
        elsif line =~ /transitive\s*\[\s*(\w+)\s*,\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "transitive" + " " + $1 + " " + $2 + " " + $3
        elsif line =~ /noOrphans\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "noOrphans" + " " + $1 + " " + $2
            delclass = $1
        elsif line =~ /noDangling\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "noDangling" + " " + $1 + " " + $2
            delclass = $1
        elsif line =~ /deletePropagates\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "deletePropagates" + " " + $1 + " " + $2
            delclass = $1
        #elsif line =~ /deletePropagates2\s*\[\s*(\w+)\s*,\s*(\w+)\s*,\s*(\w+)\s*\]/ 
        #	property = "deletePropagates2" + " " + $1 + " " + $2 + " " + $3
        #	delclass = $1
        elsif line =~ /noDeletePropagation\s*\[\s*(\w+)\s*,\s*(\w+)\s*\]/ 
            property = "noDeletePropagation" + " " + $1 + " " + $2
            delclass = $1
        #elsif line =~ /noDeletePropagation2\s*\[\s*(\w+)\s*,\s*(\w+)\s*,\s*(\w+)\s*\]/ 
        #	property = "noDeletePropagation2" + " " + $1 + " " + $2 + " " + $3
        #	delclass = $1
        end

        if property.nil?
            puts "Property on line #{l} not recognized.  Check syntax and try again."
            exit
        else
            properties << [property, line, delclass, predicate, static]
        end
    end
    
    return properties
end


# Times how long a block of code takes to run
def timer  
    start = Time.now  
    yield  
    return Time.now - start
end

# Returns OS-specific dev/null string
def blackhole
    nullfile = "/dev/null" 
    if RUBY_PLATFORM =~ /mswin|mingw/
        nullfile = "NUL"
    end
    return nullfile
end

# Performs bounded verification using the Alloy Analyzer
# - 'model_directory': location of directory containing Rails data models to verify
# - 'properties': array of property strings to be verified, e.g. [ "alwaysRelated Book User", ...]
# - 'do_csv': true if output format should be CSV
def verification_alloy( model_directory, properties, bound, do_csv = false )
  csv = CSVHashFormatter.new if do_csv
    properties.each do |prop|
        alloyoutput = nil
        alloytime = nil
        alloyvars = nil
        alloyclauses = nil
        verif_result = nil
    
    totaltime = timer do
      # Translate RoR data models to Alloy
      translate_cmd = "ruby activerecord2alloy_dynamic.rb  -p \"#{prop[0]}\"  -b \"#{bound}\"  -o #{ALLOYSPEC} #{model_directory}"
      translate_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings to file; right now just ignoring\hiding
      translate = %x[#{translate_cmd}]
 
      if not translate.empty?
        if do_csv
          # do not abort in this mode, instead report a translator error and continue to the next property
          alloyvars = ''
          alloyclauses = ''
          alloytime = 0
          verif_result = :error
          next
        else
          print "---Translator error:\n #{translate}"   # TODO: put in log, and don't abort
          exit
        end
      end
      
      #Run spec through Alloy Analyzer for verification
      alloytime = timer do
        alloyoutput = %x[java -cp .#{File::SEPARATOR}AlloyExecuter.jar AlloyExecuter  #{ALLOYSPEC}]	
      end
      ###puts alloyoutput

      #Parse output and determine satisfiability, and the number of vars/clauses
      outputlines = alloyoutput.lines
      begin
        varscls = outputlines.next
        if varscls =~ /^Vars: (\d+) Clauses: (\d+)/
          alloyvars = $1
          alloyclauses = $2
        else
          raise 'vars/clauses line not found'
        end
        
        satisf = outputlines.next
        if satisf =~ /^Unsatisfiable/
          verif_result = :unsat
        elsif satisf =~ /^Satisfiable/
          verif_result = :sat
        else
          raise 'satisfiability result not found'
        end
      rescue Exception => e
        if do_csv
          # just continue on to the next property
          alloyvars    ||= ''
          alloyclauses ||= ''
          alloytime    ||= 0
          verif_result = :error
        else
          puts "--Error processing Alloy output for property #{prop[1].strip}."
          ###TODO: put Alloy output in log, and contents of "e"
          exit
        end
      end
    
    end # getting totaltime
    
    #Print information about verification and results
    if do_csv       
      print_verif_stats_csv(csv, prop[1].strip, prop[3], "Alloy", alloyvars, 
        alloyclauses, alloytime, totaltime, verif_result)
    else
      print_property_info( prop[1], false, "Alloy", alloyvars, 
        alloyclauses, alloytime, totaltime )
      
      print_instance = print_verif_result( verif_result, prop[3], bound, true )
      
      print_alloy_instance( alloyoutput, prop[4] ) if print_instance
    end
  end
  puts csv if do_csv
end


# Processes and prints instance in our formal language
#  - 'alloy_output': output of Alloy Analyzer
#  - 'isStaticProp': true if this is an instance of a static property
#	and thus PostState info should not be printed; false otherwise
def print_alloy_instance( alloy_output, isStaticProp)
    # Variables
    specfile = nil	# pointer to alloy spec file
    types = []		# list of types, e.g. ["Book", "User", ...]
    plural_types = {}# mapping between pluralized version of a type (found 
                #  in instance) and type (found in 'types' array)
                #  e.g. "books" -> "Book"
    pre_instances = {}	# mapping from type to instances in PreState, 
                    #  e.g. "Book" -> ["Book$0", "Book$1", ...]
    pre_relations = {}# maps relations in instance tuples in PreState, e.g.
                 #  "user_books" -> [["User$1", "Book$0"], ...]
    post_instances = {}	# instances in the PostState
    post_relations = {}	# relations in the PostState
                
                
    # Get sigs from spec
    begin
         specfile = File.open( ALLOYSPEC, "r" )
    rescue SystemCallError => e
        puts "--Error opening internal file: #{e}"
        exit
    end

    specfile.each do |line| 
        if line =~ /^sig (\w+) extends ActiveRecord/ # all sigs
            types << $1
        elsif line =~ /^sig (\w+) in (\w+)/  # Conditional sigs
            types << $1 if $2 != "ActiveRecord"
        elsif line =~ /^one sig PreState/  # All sigs at beginning of file
            break
        end
    end
    specfile.close
    
    
    # Create pluralized versions of sigs (as found in instance)
    types.each do |t|
        plural_types[t.downcase.pluralize] = t
    end
    
    
    # Parse instance to get object (type) and relation instances
    prefixlen = "PreState$0->".length
    postfixlen = "PostState$0->".length
    alloy_output.each_line do |line|
        if line =~ /^this\/PreState<:(.+)=\{(.+)\}/
            instances = $2[prefixlen, $2.length - prefixlen]
            instances = instances.split(", PreState$0->")
            type = plural_types[$1]
            
            if type  #object instances	
                pre_instances[type] = instances
            elsif instances[0].include?("->") #relation instances
                pre_relations[$1] = []
                instances.each do |i|
                    pre_relations[$1] << i.split("->")
                end
            end
        elsif not isStaticProp and line =~ /^this\/PostState<:(.+)=\{(.+)\}/
            instances = $2[postfixlen, $2.length - postfixlen]
            instances = instances.split(", PostState$0->")
            rel = $1[0, $1.length - 1]
            type = plural_types[ rel ]

            if type #object instances
                post_instances[type] = instances
            elsif instances[0].include?("->") #relation instances
                post_relations[rel] = []
                instances.each do |i|
                    post_relations[rel] << i.split("->")
                end
            end			
        end
    end

    
    # Print objects as   Classname { o1, o2, o3 }
    print "\nOBJECTS:\n"
    pre_instances.each do |klass, objs|
        print "  ", klass, " { "
        numobjs = objs.length - 1
        objs[0, numobjs].each {|o| print o, ", "}
        print objs[numobjs], " }\n"
    end
    puts
    post_instances.each do |klass, objs|
        print "  Post_", klass, " { "
        numobjs = objs.length - 1
        objs[0, numobjs].each {|o| print o, ", "}
        print objs[numobjs], " }\n"
    end

    # Print relations as   Relname { (a1, b2), (a2, b3) }
    print "\nRELATIONS:\n"
    pre_relations.each do |relname, objs|
        print "  ", relname, " { "
        numobjs = objs.length - 1
        objs[0, numobjs].each { |o| print "(", o[0], ", ", o[1] , "), " }
        print "(", objs[numobjs][0], ", ", objs[numobjs][1], ") }\n"
    end
    puts
    post_relations.each do |relname, objs|
        print "  Post_", relname, " { "
        numobjs = objs.length - 1
        objs[0, numobjs].each { |o| print "(", o[0], ", ", o[1] , "), " }
        print "(", objs[numobjs][0], ", ", objs[numobjs][1], ") }\n"
    end
end


# Perform automatic property inference and unbounded verification
# - 'model_directory': location of directory containing Rails data models to verify
# - 'projection_on': true if projection should be performed; false otherwise
def property_inference( model_directory, projection_on )
    inference_cmd = "ruby property_inferer.rb #{model_directory}"
    inference_cmd << " -j " if projection_on
    inference_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings
    exec( inference_cmd )
end

# Create and output an entity-relation diagram(ERD), as a png or dot file
# - 'model_directory': location of Rails data models from which to create ERD
# - 'filename': The name of the file to output the ERD to.  
#    If 'filename' ends in .png, the output will be a png file; otherwise it will be a .dot file
def create_erd( model_directory, out_filename )
    start_directory = Dir.getwd 

    begin
        Dir.chdir(model_directory)
    rescue SystemCallError => e
		print "Error navigating to directory: " + e.to_s + "\n"
		exit
    end
	 
	 Dir.foreach(".") do |filename| 
		if filename =~ /\.rb$/ and File.file?(filename) then
			# Get contents of model files
			file = File.open(filename, 'r')
			file_contents = ""
			file.each {|line| file_contents << line }
			file.close

			# Parse
			s_exp = RubyParser.new.parse file_contents			
			ActiveRecordProcessor.new.process(s_exp)
		end
	end
	Dir.chdir start_directory
	
	
	# Create ERD
    output_as_png = out_filename.end_with? ".png"
    ActiveRecordProcessor.create_erd( out_filename, output_as_png ) 
end

# Performs unbounded verification using Z3
# - 'model_directory': location of directory containing Rails data models to verify
# - 'properties': array of property strings to be verified, e.g. [ "alwaysRelated Book User", ...]
# - 'projection_on': true if projection should be performed; false otherwise
# - 'do_csv': true if output format should be CSV
def verification_z3( model_directory, properties, projection_on, do_csv = false )
    csv = CSVHashFormatter.new if do_csv
    properties.each do |prop|  # Verify properties one by one
        # Variables
        $relations = {}  	 # relations["relation_name"] = RelationInfo object
                             # (contains only those relations defined in Z3 instance)
        $post_relations = {} # post_relations["relation_name"] = ["type1", "type2"]
                             #   i.e. stores the argument types of a Post_relation function
        $post_objects = {}   # post_objects["relation_name"] = "type1"
                             #   i.e. stores the argument type of a Post_object function
        specfile = nil	     # pointer to generated SMT spec file
        instancefile = nil	 # file to store instance for processing (and printing in formal lang)
        delobjarg = ""       # (for delete properties) the class whose object to delete
        smtvars = nil	     # number of variables in SMT spec
        smtclauses = nil	 # number of clauses in SMT spec
        z3output = ""	     # everything that z3 outputs
        z3time = 0		     # time Z3 takes to perform verification
        objects = {} 	     # objects["Classname"] = ["Obj1", "Obj2", ...]
        poly_fxns = []          # e.g. ["isBlog", "isPhoto", "isProfile"]
        verif_result = nil	 # After verification this should hold :sat :unsat :timeout or :error

    totaltime = timer do
      # Translate RoR data models to SMT
      delobjarg = " -d #{prop[2]} " if prop[2]
      projection = " -j " if projection_on
      translate_cmd = "ruby activerecord2SMT_prop.rb #{model_directory} -p \"#{prop[0]}\" #{projection} #{delobjarg} -o #{SMTSPEC}"
      translate_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings to file; right now just ignoring\hiding
      translate = %x[#{translate_cmd}]
        
      if not translate.empty?
        print "---Translator error:\n #{translate}"   # TODO: put in log, and don't abort
        exit
      end
      
      # Get number of variables and clauses produced
      stats = %x[ruby stats_smtlibspec.rb #{SMTSPEC} ]
      
      if stats =~ /TOTAL variables: (\d+)/
        smtvars = $1
      end 
      if stats =~ /TOTAL clauses: (\d+)/
        smtclauses = $1
      end
      
      # Run SMT spec through Z3 for verification
      z3time = timer do
        z3_cmd = ".\\z3.exe  -T:10 -smt2 #{SMTSPEC} "
        z3_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings to file; right now just ignoring\hiding
        z3output = %x[#{z3_cmd}]
      end
      ##print z3output; exit  # for debugging
        
    end # getting total verification time

        # Check to make sure didn't get any errors when running Z3
        if not z3output =~ /^unsat/ and z3output =~ /\(error/                            
            verif_result = :error
            
            ## TODO: Output the following to the log (the errors output by Z3) and 
            ##       point to smtspec location
              errors = ""
              z3output.each_line do |line|
                if line =~/^sat/
                  break
                else
                  errors << line
                end
              end
              
              puts "---Verification error:\n#{errors}"
             # puts "\nPlease fix the error(s) in the SMT spec: #{SMTSPEC}"
        elsif z3output.scan(/\(error/).size > 1
            verif_result = :error        
            
            ## TODO: log error
            
        else
            #PROCESS SMT SPEC TO PREPARE FOR PRINTING INSTANCE:
            #--------------------------------------------------
            
            # Get sorts from SMT spec
            begin
                specfile = File.open( SMTSPEC, "r" )
                instancefile = File.new( INSTANCE, "w" )
            rescue SystemCallError => e
                puts "--Error opening internal file: #{e}"
                puts Dir.pwd
                exit
            end
            #instancefile.puts   "(set-option :model_completion true)\n\n"

            get_poly_subclasses = false
            specfile.each do |line| 
                if get_poly_subclasses
                    # from line like "	(or (isBlog p) (isPhoto p) (isProfile p) )" obtains
                    #  "isBlog, isPhoto, isProfile" --> the polymorphic subclasses
                    line.strip!
                    line = line[5, line.length - 10]
                    poly_fxns = line.split(" p) (")
                    get_poly_subclasses = false
                elsif line =~ /^\(assert \(forall \(\(p PolymorphicClass\)\)/
                    get_poly_subclasses = true
                elsif line =~ /\(declare-sort ([^0]+) /
                    objects[$1] = []
                    instancefile.puts line
                end
            end
            specfile.close


            # Get object and relation names from Z3 instance (removing quantifiers 
            #  because there is a bug in those lines in Z3 output as confirmed by Microsoft)
            #  Update: Bug fixed in version 4.3 of Z3
            # At the same time, determine verification result (sat, unsat, or timeout)
            copy_rest = false
            z3output.each_line do |line|
                if copy_rest and not line =~ /^\)/
                    helper_get_relation_from line
                    instancefile.puts line
                elsif line =~ /^sat/
                    verif_result = :sat
                elsif line =~ /^unsat/
                    verif_result = :unsat
                elsif line =~ /^timeout/
                    verif_result = :timeout
                elsif line =~ /\(declare-fun ([^(]+) \(\) ([^)]+)/
                    instancefile.puts line
                    objects[$2] << $1
                elsif line =~ /\(define-fun/
                    instancefile.puts line
                    helper_get_relation_from line
                    copy_rest = true
                end
            end
            instancefile.puts "(check-sat)"
            instancefile.puts "(get-model)"
            instancefile.close

        end
        
        #Print information about verification and results
        if do_csv
            print_verif_stats_csv(csv, prop[1].strip, prop[3], "Z3", smtvars, smtclauses, z3time, totaltime, verif_result)
        else
            if verif_result.nil? 
                puts "--Error processing Z3 instance."
                exit
            end
            
            print_property_info( prop[1], true, "Z3", smtvars, smtclauses, z3time, totaltime )
            print_instance = print_verif_result( verif_result, prop[3], nil, true )
            print_z3_instance( objects, poly_fxns ) if print_instance
        end
    
    end  #end for-each property
  
    puts csv if do_csv
end


# Helper function to verification_z3(): 
#  From a "(define-fun" line in a Z3 instance, determines whether it defines a relation
#  If so, gets the relation signature and adds it to $relations, $post_relations or $post_objects
#  (these variables initialized in verification_z3() and used in print_z3_instance())
def helper_get_relation_from (line)
    # Ignore polymorphic functions
    if line =~ /\(define-fun is([^(!]+) \(\(\S+ PolymorphicClass\)\) Bool/

    # Two argument definitions  (for many-to-many or Post_relation functions)
    elsif line =~ /\(define-fun ([^(!]+) \((\(\S+ \S+\)) (\(\S+ \S+\))\) ([a-zA-Z]+)/
        rel_name = $1
        arg1 = $2
        arg2 = $3
        return_type = $4
        
        arg1.match( /\(([^)]+) ([^)]+)\)/ )
        arg1_type = $2
        arg2.match( /\(([^)]+) ([^)]+)\)/ )
        arg2_type = $2
        
        if not rel_name.start_with?("Post_")
            relation = RelationInfo.new(2, [arg1_type, arg2_type], return_type)
            $relations[rel_name] = relation
        else
            $post_relations[rel_name] = [arg1_type, arg2_type]
        end
        
    # One argument definitions (for all other functions (one-to-one, one-to-many
    #   relations, subclass relations, functions, Post_object functions))
    elsif line =~ /\(define-fun ([^(!]+) \((\(\S+ \S+\))\) ([a-zA-Z]+)/
        rel_name = $1
        args = $2
        return_type = $3
        
        args.match( /\(([^)]+) ([^)]+)\)/ )
        arg_type = $2
        
        if not rel_name.start_with?("Post_")
            relation = RelationInfo.new(1, [arg_type], return_type)
            $relations[rel_name] = relation
        else
            $post_objects[rel_name] = arg_type
        end
    end
end


# Processes and prints instance in our formal language
#  - 'objects': a hash that maps a class to its instances:
#		      objects["Classname"] = ["Obj1", "Obj2", ...]
#  - 'poly_fxns': an array of polymorphic function names in an SMT spec, 
#		      e.g. ["isBlog", "isPhoto", "isProfile"]
def print_z3_instance( objects, poly_fxns )
    # Variables
    relationsfile = nil		 # the instance file from verification_z3() appended with eval statements
                     #  to determine which objects are related together and which objects 
                     #  and relations are in the Post state
    z3output = ""	     	 # everything that z3 outputs

    
    # Setup: Find which objects are related to which by creating a
    #  file of relations + values we're interested in evaluating
    begin
        relationsfile = File.open( INSTANCE, "a" )
    rescue SystemCallError => e
        puts "--Error creating internal file: #{e}"
        exit
    end

    poly_fxns.each do |relname|
        objects["PolymorphicClass"].each do |o1|
            relationsfile.print "(eval (", relname, " ", o1, ") :model-completion true)\n"
            sort = relname.gsub(/^is/, '')
        end
    end
    
    $post_objects.each do |relname, type|
        objects[type].each do |o|
            relationsfile.print "(eval (", relname, " ", o, ") :model-completion true)\n"
        end
    end

    $relations.each do |relname, info|
        if info[:num_args] == 2
            objects[ info[:arg_types][0] ].each do |o1|
                objects[ info[:arg_types][1] ].each do |o2|
                    relationsfile.print "(eval (", relname, " ", o1, " ", o2, ") :model-completion true)\n"
                end
            end
            
        else
            objects[ info[:arg_types][0] ].each do |o1|
                relationsfile.print "(eval (", relname, " ", o1, ") :model-completion true)\n"
            end
        end
    end
    
    $post_relations.each do |relname, types|
        objects[ types[0] ].each do |o1|
            objects[ types[1] ].each do |o2|
                relationsfile.print "(eval (", relname, " ", o1, " ", o2, ") :model-completion true)\n"
            end
        end		
    end
    relationsfile.close

 
    # Send 'relationsfile' to Z3 for evaluation
    z3_cmd = ".\\z3.exe  -T:10 -smt2 #{INSTANCE} "
    z3_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings to file; right now just ignoring\hiding
    z3output = %x[#{z3_cmd}]
    
    # check to make sure didn't get any errors when running Z3
    if not z3output =~ /^unsat/ and z3output =~ /\(error/ 
        print "\nError forming instance.\n"		
    end
    
    # Remove any output that doesn't correspond to result of an eval  (e.g. let statements)
    cleaned_output = ""
    z3output.each_line do |line|
        cleaned_output << line if line =~ /^[a-zA-Z]/
    end
    
    
    # Parse output and print instance
    evals = cleaned_output.lines
    evals.next  # removes first line that says "sat"
    
    # Print objects as   Classname { o1, o2, o3 }
    print "\nOBJECTS:\n"
    objects.each do |klass, objs|
        if objs.length > 0 and klass != "PolymorphicClass"
            print "  ", klass, " { "
            objstring = ""
            objs.each {|o| objstring << o << ", "}
            print objstring.sub(/, $/, ' '), "}\n"
        end
    end
    
    poly_fxns.each do |relname|
        klass = relname.gsub(/^is/, '')
        objstring = ""
        objects["PolymorphicClass"].each do |o|
            objstring << o << ", " if evals.next == "true\n"
        end
        if not objstring.empty?
            print "  ", klass, " { "
            print objstring.sub(/, $/, ' '), "}\n" 
        end
    end
    
    $post_objects.each do |klass, type|
        objstring = ""
        objects[type].each do |o|
            objstring << o << ", " if evals.next == "true\n"
        end
        if not objstring.empty?
            print "  ", klass, " { "
            print objstring.sub(/, $/, ' '), "}\n" 
        end
    end

    
    # Print relations as   Relname { (a1, b2), (a2, b3) }
    print "\nRELATIONS:\n"
    $relations.each do |relname, info|
        relatedobjs = ""
        
        if info[:num_args] == 2
            objects[ info[:arg_types][0] ].each do |o1|
                objects[ info[:arg_types][1] ].each do |o2|
                    relatedobjs << "(" << o1 << ", " << o2 << "), "  if evals.next == "true\n"
                end
            end
        else
            objects[ info[:arg_types][0] ].each do |o1|
                o2 = evals.next
                o2.rstrip!
                relatedobjs << "(" << o1 << ", " << o2 << "), "
            end
        end
        
        if not relatedobjs.empty?
            print "  ", relname, " { "
            print relatedobjs.sub(/, $/, ' '), "}\n"
        end
    end
    
    $post_relations.each do |relname, types|
        relatedobjs = ""
        objects[ types[0] ].each do |o1|
            objects[ types[1] ].each do |o2|
                relatedobjs << "(" << o1 << ", " << o2 << "), "  if evals.next == "true\n"
            end
        end	
        
        if not relatedobjs.empty?
            print "  ", relname, " { "
            print relatedobjs.sub(/, $/, ' '), "}\n"
        end	
    end

end #end-fxn print_z3_instance()




# Performs unbounded verification using Spass
# - 'model_directory': location of directory containing Rails data models to verify
# - 'properties': array of property strings to be verified, e.g. [ "alwaysRelated Book User", ...]
# - 'do_csv': true if output format should be CSV
def verification_spass( model_directory, properties, do_csv = false )
  csv = CSVHashFormatter.new if do_csv
  properties.each do |prop|   # Verify properties one by one
    prop_type = prop[0].split[0].to_sym
    
    # Variables
    verif_result = :timeout # Indicates result of verification.  Can only be one of the 
            #  following values-- :timeout, :unsat, or :sat
    num_vars = '-'    # Number of variables in the Spass specification
    num_clauses = nil # Number of clauses in the Spass specification
    verif_time = nil    # Time the solver took to perform the verification
    total_time = nil    # Total verification time (including translation of 
            #  Active Record files into first-order logic, etc)
 
    # to make a result inverted for a given property pattern just add the symbol of the property type to 
    # this set
    properties_for_which_satisfiable_is_good = Set[]

    # Verify property
    temp = nil
    output = nil
    total_time = timer do
      begin
        command = "ruby activerecord2spass.rb -p \"#{prop[0]}\" #{model_directory}"
        output = `#{command}`
        
          temp = Tempfile.new 'spass_translation'
        temp.write output
        temp.close
        
        if $?.exitstatus != 0
          verif_result = :error
          next
        end

        output = `#{OS.spass_executable} -TimeLimit=20 #{temp.path}`
        
        if $?.exitstatus != 0
          verif_result = :error
          next
        end
        
        output = output.split('\n')
        
        output.select{ |line| line =~ /^\s*Axiom clauses: (\d+) Conjecture clauses: (\d+)\s*$/ }.each do |line|
          num_clauses = $1 + $2
        end
        verif_time = 0
        output.select{ |line| line =~ /(\d+):(\d\d):(\d\d)\.(\d\d)/ }.each do |line|
          verif_time += 0.1*$4.to_i + $3.to_i + 60*$2.to_i + 60*60*$1.to_i
        end

        proven = output.select{ |line| line =~ /.*Proof found.\s*$/}.length > 0
        not_proven = output.select{ |line| line =~ /.*Completion found.\s*$/}.length > 0

        if proven or not_proven # otherwise, it's a timeout
          if properties_for_which_satisfiable_is_good.include?(prop_type)
            verif_result = proven ? :sat : :unsat
          else
            verif_result = proven ? :unsat : :sat
          end
        end
      rescue SystemExit, Interrupt
        puts output
        raise
      ensure
        temp.delete unless temp.nil?
      end
    end
    
    if do_csv
      print_verif_stats_csv(csv, prop[1].strip, prop[3], 'Spass', num_vars, num_clauses, verif_time, total_time, verif_result)
    else
      # Print information about verification and results
      print_property_info( prop[1], true, "Spass", num_vars, num_clauses, verif_time, total_time )
      print_verif_result( verif_result, prop[3], nil, false )
    end
  end
  puts csv if do_csv
end


# Prints information about a property being verified
#  - 'property': text of the property that is being verified (e.g. "alwaysRelated[ Book, User ]" (string)
#  - 'unbounded': true if unbounded verification is being performed; false otherwise (boolean)
#  - 'solver': solver being used (e.g. "Alloy") (string)
#  - 'num_vars': the number of variables created by the solver (int)
#  - 'num_clauses': the number of clauses created by the solver (int)
#  - 'verif_time': time the solver took to perform the verification (float)
#  - 'total_time': total verification time (float)
def print_property_info( property, unbounded, solver, num_vars, num_clauses, verif_time, total_time )
    puts "\n\n--------"
    puts "Property being verified: " + property
    puts "--------"
    puts "Verification technique: " + (unbounded ? "Unbounded" : "Bounded")
    puts "Solver used: #{solver}"
    print "Formula size: "
    if num_vars and num_clauses
        puts "#{num_vars} variables, #{num_clauses} clauses"
    else
        puts "Not provided"
    end
    puts "Verification time: #{verif_time ? verif_time.round(3) : 0}"
    puts "Total time: #{total_time.round(3)}"
    puts "------"
end	


# Prints verification results uniformly.  Returns true if an instance needs to be printed 
# and false otherwise.
#  - 'verif_result': :timeout, :unsat, or :sat
#  - 'isPredicateProperty': true if property that was verified was a predicate; false if
#		property was an assert
#  - 'bound': the bound if bounded verification was performed; nil to indicate 
#		unbounded verification
#  - 'instance_produced': boolean variable indicating whether the solver produced instances
def print_verif_result( verif_result, isPredicateProperty, bound, instance_produced )
    print_property = false	
    
    if verif_result == :timeout
        puts "Result: Timeout. Property could not be verified."  
            #TODO: "Use the -t option to increase the timeout length if desired."
    elsif verif_result == :error
        puts "Result: There was an error with the solver."
    elsif isPredicateProperty
        if verif_result == :unsat
            if bound.nil?
                no_instance = instance_produced ? "  (No satisfying instance found.)" : ""
                puts "Result: Property does not hold.#{no_instance}"
            else
                puts "Result: Property may not hold. (No satisfying instance found within bound of #{bound}.)"
            end
        elsif verif_result == :sat
            instance_below = instance_produced ? "  A satisfying instance is displayed below." : ""
            puts "Result: Property holds.#{instance_below}"
            print_property = true
        end
    else  # Assertion property
        if verif_result == :sat
            instance_below = instance_produced ? "  A counterexample instance is displayed below." : ""
            puts "Result: Property does not hold.#{instance_below}"
            print_property = true
        elsif verif_result == :unsat
            if bound.nil?
                puts "Result: Property holds."
            else
                puts "Result: Property may hold. (No counterexample found within bound of #{bound}.)"
            end
        end
    end
    puts "------"
    
    return print_property
end

def print_verif_stats_csv( csv, property, isPredicateProperty, solver, num_vars, num_clauses, verif_time, total_time, verif_result )
    result = verif_result    
    result = :error if verif_result.nil?
    if isPredicateProperty
        if verif_result == :sat
            result = true
        elsif verif_result == :unsat
            result = false
        end
    else
        if verif_result == :sat
            result = false
        elsif verif_result == :unsat
            result = true
        end
    end
        
  csv << {
    :property    => property,
    :solver      => solver,
    :num_vars    => num_vars,
    :num_clauses => num_clauses,
    :verif_time  => verif_time,
    :total_time  => total_time,
    :result      => result
  }
end

#--------------------------------------------------#
#------------------ Start script --------------------#
#--------------------------------------------------#

# Variables
options = {}           # used to define the command-line options
properties = []        # property strings extracted from input file
model_directory = nil  # location and name of directory containing the 
                       # Ruby on Rails application's model files


# Set up OptionParser (to parse command-line arguments)
OptionParser.new do |opts|
    opts.banner = "USAGE: ruby iDaVeR.rb [options] path\\to\\models\\folder" +
        "\n  This program verifies the specified properties about a set of Ruby on Rails" +
        "\n  data models using the technique specified. (Bounded verification performed" +
        "\n  by default.) If no properties are specified using the -p option, automatic" +
        "\n  property inference is performed. Inferred properties are automatically" +
        "\n  verified using unbounded verification."

     
    # Define the options
    options[:propertyfile] = nil
    opts.on( '-p', '--property FILENAME',
       'Name of the file that contains the properties to verify' ) do |propfile|
        options[:propertyfile] = propfile
    end
    
    #Take out Spass as an option for an unbounded solver:
    #options[:unbounded] = nil
    #opts.on( '-u', '--unbounded_verif SOLVER',
    #  'Performs unbounded rather than bounded verification using the solver specified (z3 or sp, for Z3 or Spass)' ) do |s|
    #	options[:unbounded] = s
    #end
    
    options[:unbounded] = nil
    opts.on( '-u', '--unbounded_verif',
      'Performs unbounded (rather than bounded) verification using the SMT Solver, Z3' ) do
        options[:unbounded] = true
    end
  
    options[:bound] = DEFAULT_BOUND
    opts.on( '-b', '--bound INT', Integer, 
      "Uses given bound rather than default of #{DEFAULT_BOUND} (bounded verification only)" ) do |b|
        options[:bound] = b
    end	
    
    options[:projection] = nil
    opts.on( '-j', '--projection_on',
      'Performs data model projection during verification (z3 only)' ) do 
        options[:projection] = true
    end	
  
    options[:csv] = false
    opts.on( '-c', '--csv', 
        "Output verification stats as CSV. Default: #{options[:csv]}") do
        options[:csv] = true
    end
    
    options[:erd] = nil
    opts.on( '-e', '--erd [FILENAME]', 
      "Outputs entity-relation diagram (ERD) for given data model. " + 
      "(All other options ignored.) " + 
      "Output as a .dot file unless user-specified FILENAME ends with .png. " + 
      " Default output filename: #{DEFAULT_ERD_FILENAME}" ) do |f|
        options[:erd] = f || DEFAULT_ERD_FILENAME
    end
    
    # This displays the help screen
    opts.on( '-h', 'Display this screen' ) do
        puts opts
        exit
    end
end.parse!  # parse options off of command line



# Get models folder
if ARGV.length == 1 	    # model directory should be the only thing left on the command line
    model_directory = ARGV[0]
else
    puts "  Program invoked incorrectly.  Use '-h' option for usage information."
    exit
end

if options[:erd]  # Output ERD
    create_erd( model_directory, options[:erd] )
    
elsif options[:propertyfile].nil?  # Perform automatic property inference
    property_inference( model_directory, options[:projection] )

else  # Perform verification (using specified technique) on user-input properties
    
    # Obtain properties from file
    properties = get_properties( options[:propertyfile] )

    if properties.empty?
        puts "  File provides no properties to check.  Exiting."
        exit    
    end

    # Perform verification
    if options[:unbounded] 
        #if options[:unbounded] == "z3"
            verification_z3( model_directory, properties, options[:projection], options[:csv] )
      #Took out Spass as an option for an unbounded solver:
        #elsif options[:unbounded] == "sp"
        #	verification_spass( model_directory, properties, options[:csv] )
        #else
        #	puts "  Error: Unbounded solver not recognized.  Use either 'z3' or 'sp'."
        #	exit
        #end
    else 
        verification_alloy( model_directory, properties, options[:bound], options[:csv] )
    end
end